import operator
from typing import Optional, Dict, Tuple, List, Union

import pandas as pd
import pytest
import sklearn
import numpy as np
import plotly.graph_objs as go

from trelawney.base_explainer import BaseExplainer


class FakeExplainer(BaseExplainer):

    def fit(self, model: sklearn.base.BaseEstimator, x_train: pd.DataFrame, y_train: pd.DataFrame):
        return super().fit(model, x_train, y_train)

    @staticmethod
    def _regularize(importance_dict: List[Tuple[str, float]]) -> List[Tuple[str, float]]:
        total = sum(map(operator.itemgetter(1), importance_dict))
        return [
            (key, -(2 * (i % 2) - 1) * (value / total))
            for i, (key, value) in enumerate(importance_dict)
        ]

    def feature_importance(self, x_explain: pd.DataFrame, n_cols: Optional[int] = None):
        importance = self._regularize(sorted(
            ((col, np.mean(np.abs(x_explain.loc[:, col]))) for col in x_explain.columns),
            key=operator.itemgetter(1),
            reverse=True
        ))
        total_mvmt = sum(map(operator.itemgetter(1), importance))
        res = dict(importance[:n_cols])
        res['rest'] = total_mvmt - sum(res.values())
        return res

    def explain_local(self, x_explain: pd.DataFrame, n_cols: Optional[int] = None):
        res = []
        for sample_explanation in x_explain.abs().to_dict(orient='records'):
            importance = self._regularize(sorted(sample_explanation.items(), key=operator.itemgetter(1), reverse=True))

            total_mvmt = sum(map(operator.itemgetter(1), importance))
            res_ind = dict(importance[:n_cols])
            res_ind['rest'] = total_mvmt - sum(res_ind.values())
            res.append(res_ind)

        return res


def _float_error_resilient_compare(left: Union[List[Dict], Dict], right: Union[List[Dict], Dict]):
    assert len(left) == len(right)
    if isinstance(left, list):
        return [_float_error_resilient_compare(ind_left, ind_right) for ind_left, ind_right in zip(left, right)]
    for key, value in left.items():
        assert key in right
        assert abs(value - right[key]) < 0.0001


def test_explainer_basic():

    explainer = FakeExplainer()
    _float_error_resilient_compare(
        explainer.feature_importance(pd.DataFrame([[10, 0], [0, -5]], columns=['var_1', 'var_2'])),
        {'var_1': 5 / 7.5, 'var_2': -2.5 / 7.5, 'rest': 0.}
    )
    _float_error_resilient_compare(
        explainer.feature_importance(pd.DataFrame([[10, 0], [0, -5]], columns=['var_1', 'var_2']), n_cols=1),
        {'var_1': 5. / 7.5, 'rest':  -2.5 / 7.5}
    )

    _float_error_resilient_compare(
        explainer.explain_local(pd.DataFrame([[10, 0], [0, -5]], columns=['var_1', 'var_2'])),
        [{'var_1': 1., 'var_2': 0., 'rest': 0.}, {'var_2': 1., 'var_1': 0., 'rest': 0.}]
    )

    _float_error_resilient_compare(
        explainer.explain_local(pd.DataFrame([[10, 0], [0, -5]], columns=['var_1', 'var_2']), n_cols=1),
        [{'var_1': 1., 'rest': 0.},{'var_2': 1, 'rest': 0.}]
    )


def test_explainer_filter():

    explainer = FakeExplainer()
    _float_error_resilient_compare(
        explainer.filtered_feature_importance(pd.DataFrame(
            [[10, 0, 4], [0, -5, 3]], columns=['var_1', 'var_2', 'var_3']), cols=['var_1', 'var_3']
        ),
        {'var_1': 10 / 22, 'var_3': -7 / 22, 'rest': 5 / 22}
    )

    _float_error_resilient_compare(
        explainer.filtered_feature_importance(
            pd.DataFrame([[10, 0, 4], [0, -5, 3]], columns=['var_1', 'var_2', 'var_3']), n_cols=1,
            cols=['var_1', 'var_3']
        ),
        {'var_1': 10 / 22, 'rest': -2 / 22}
    )

    _float_error_resilient_compare(
        explainer.explain_filtered_local(
            pd.DataFrame([[10, 0, 4], [0, -5, 3]], columns=['var_1', 'var_2', 'var_3']), cols=['var_1', 'var_3']
        ),
        [{'var_1': 10 / 14, 'var_3': -4 / 14, 'rest': 0.}, {'var_3': -3 / 8, 'var_1': 0., 'rest': 5 / 8}]
    )

    _float_error_resilient_compare(
        explainer.explain_filtered_local(
            pd.DataFrame([[10, 0, 4], [0, -5, 3]], columns=['var_1', 'var_2', 'var_3']),
            cols=['var_1', 'var_3'], n_cols=1
        ),
        [{'var_1': 10 / 14, 'rest': -4 / 14}, {'var_3': -3 / 8, 'rest': 5 / 8}]
    )


def test_local_graph(FakeClassifier, fake_dataset):

    model = FakeClassifier()
    explainer = FakeExplainer()
    explainer.fit(model, *fake_dataset)

    with pytest.raises(ValueError):
        _ = explainer.graph_local_explanation(pd.DataFrame([[10, 30], [1, 2]], columns=['var_1', 'var_2']))

    graph = explainer.graph_local_explanation(pd.DataFrame([[10, 30]], columns=['var_1', 'var_2']))

    assert len(graph.data) == 1
    assert isinstance(graph.data[0], go.Waterfall)
    waterfall = graph.data[0]
    assert waterfall.x == ('start_value', 'var_2', 'var_1', 'rest', 'output_value')
    assert waterfall.y == (0., .75, -0.25, 0., 0.5)

    graph = explainer.graph_local_explanation(pd.DataFrame([[10, 30]], columns=['var_1', 'var_2']), n_cols=1)

    assert len(graph.data) == 1
    assert isinstance(graph.data[0], go.Waterfall)
    waterfall = graph.data[0]
    assert waterfall.x == ('start_value', 'var_2', 'rest', 'output_value')
    assert waterfall.y == (0., .75, -0.25, 0.5)


def test_local_graph_hover_text(FakeClassifier, fake_dataset):

    model = FakeClassifier()
    explainer = FakeExplainer()
    explainer.fit(model, *fake_dataset)

    graph = explainer.graph_local_explanation(pd.DataFrame([[10, 30]], columns=['var_1', 'var_2']),
                                              info_values=pd.DataFrame([['foo', 3]], columns=['var_1', 'var_2']))

    assert len(graph.data) == 1
    assert isinstance(graph.data[0], go.Waterfall)
    waterfall = graph.data[0]
    assert waterfall.hovertext == ('start value', 'var_2 = 3', 'var_1 = foo', 'rest', 'output_value = 0.5')

    graph = explainer.graph_local_explanation(pd.DataFrame([[10, 30]], columns=['var_1', 'var_2']), n_cols=1)


def test_feature_importance_graph(FakeClassifier, fake_dataset):
    model = FakeClassifier()
    explainer = FakeExplainer()
    explainer.fit(model, *fake_dataset)
    graph = explainer.graph_feature_importance(
        pd.DataFrame([[10, 0, 4], [0, -5, 3]], columns=['var_1', 'var_2', 'var_3']), n_cols=1, cols=['var_1', 'var_3']
    )
    assert len(graph.data) == 1
    assert isinstance(graph.data[0], go.Bar)
    bar_graph = graph.data[0]
    assert bar_graph.x == ('var_1', 'rest')
    assert abs(bar_graph.y[0] - 10 / 22) < 0.0001
    assert abs(bar_graph.y[1] + 2 / 22) < 0.0001


def test_mixed_imputs():
    explainer = FakeExplainer()

    assert all(explainer._get_dataframe_from_mixed_input(
        pd.DataFrame([[1, 2, 3]], columns=['foo', 'bar', 'baz'])
    ) == pd.DataFrame([[1, 2, 3]], columns=['foo', 'bar', 'baz']))

    assert all(explainer._get_dataframe_from_mixed_input(
        pd.Series([1, 2, 3], index=['foo', 'bar', 'baz'])
    ) == pd.DataFrame([[1, 2, 3]], columns=['foo', 'bar', 'baz']))

    assert all(explainer._get_dataframe_from_mixed_input(
        np.array([1, 2, 3])
    ) == pd.DataFrame([[1, 2, 3]], columns=['feature_0', 'feature_1', 'feature_2']))

    assert all(explainer._get_dataframe_from_mixed_input(
        np.array([[1, 2, 3]])
    ) == pd.DataFrame([[1, 2, 3]], columns=['feature_0', 'feature_1', 'feature_2']))



"""
Module that provides the TreeExplainer class base on the Baseexplainer class
"""
import os
import tempfile
from typing import Optional, List, Dict, Union

import numpy as np
import pandas as pd
import sklearn
from subprocess import call
from sklearn import tree

from trelawney.base_explainer import BaseExplainer


class TreeExplainer(BaseExplainer):
    """
    The TreeExplainer class is composed of 4 methods:
    - fit: get the right model
    - feature_importance (global interpretation)
    - explain_local (local interpretation, WIP)
    - plot_tree (full tree visualisation)
    """

    def __init__(self, class_names: Optional[List[str]] = None):
        """
        initialize class_names, categorical_features and model_to_explain
        """
        self.class_names = class_names
        self._model_to_explain = None
        self._feature_names = None

    def fit(self, model: sklearn.base.BaseEstimator, x_train: Union[pd.Series, pd.DataFrame, np.ndarray],
            y_train: pd.DataFrame):

        x_train = self._get_dataframe_from_mixed_input(x_train)
        self._model_to_explain = model
        self._feature_names = x_train.columns
        return self

    def feature_importance(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray], n_cols: Optional[int] = None) -> Dict[str, float]:
        """
        returns a relative importance of each feature globally as a dict.
        :param x_explain: the dataset to explain on
        :param n_cols: the maximum number of features to return
        """
        x_explain = self._get_dataframe_from_mixed_input(x_explain)
        res = dict(zip(x_explain.columns, self._model_to_explain.feature_importances_))
        return res

    def explain_local(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                      n_cols: Optional[int] = None) -> List[Dict[str, float]]:
        """
        returns local relative importance of features for a specific observation.
        :param x_explain: the dataset to explain on
        :param n_cols: the maximum number of features to return
        """
        raise NotImplementedError('no consensus on which values can explain the path followed by an observation')

    def plot_tree(self, out_path: str = './tree_viz'):
        """
        creates a png file of the tree saved in out_path

        :param out_path: the path to save the png representation of the tree to
        """
        tree.export_graphviz(self._model_to_explain, out_file=out_path + '.dot', filled=True, rounded=True,
                             special_characters=True, feature_names=self._feature_names, class_names=self.class_names)
        call(['dot', '-Tpng', out_path + '.dot', '-o', out_path + '.png', '-Gdpi=600'])



import pandas as pd
import numpy as np
import pytest
from keras import layers, models
from keras.wrappers.scikit_learn import KerasClassifier
from sklearn import base
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier


@pytest.fixture
def fake_dataset():
    return (pd.DataFrame([list(range(100)), np.random.normal(size=100).tolist()], index=['real', 'fake']).T,
            (np.array(range(100)) > 50).astype(np.int16))


@pytest.fixture
def fitted_logistic_regression(fake_dataset):
    model = LogisticRegression()
    return model.fit(*fake_dataset)


@pytest.fixture
def fitted_neural_network(fake_dataset):

    def make_neural_network():
        model = models.Sequential([
            layers.Dense(2, input_shape=(2,), activation='softmax')
        ])
        model.compile(loss='categorical_crossentropy', optimizer='adam')
        return model

    model = KerasClassifier(make_neural_network, epochs=100, batch_size=100)
    model.fit(*fake_dataset)
    return model


@pytest.fixture
def FakeClassifier():

    class FakeClassifierInner(base.BaseEstimator):

        def fit(self, X, y):
            return self

        def predict_proba(self, X):
            return np.array([[0.5, 0.5] for _ in  range(X.shape[0])])

        def predict(self, X):
            return self.predict_proba(X)[:, 1] >= 0.5

    return FakeClassifierInner


@pytest.fixture
def fitted_decision_tree(fake_dataset):

    model = DecisionTreeClassifier()
    model.fit(*fake_dataset)
    return model



import numpy as np
import pandas as pd
import pytest

from xgboost import XGBClassifier

from trelawney.lime_explainer import LimeExplainer


def _do_explainer_test(explainer, data_to_test=None, col_real='real', col_fake='fake'):
    data_to_test = data_to_test if data_to_test is not None else pd.DataFrame([[5, 0.1], [95, -0.5]], columns=[col_real, col_fake])
    explanation = explainer.explain_local(data_to_test)
    assert len(explanation) == data_to_test.shape[0] if len(data_to_test.shape) == 2 else 1
    for single_explanation in explanation:
        assert abs(single_explanation[col_real]) > abs(single_explanation[col_fake])


def test_lime_explainer_single(fake_dataset, fitted_logistic_regression):
    explainer = LimeExplainer(class_names=['false', 'true'])
    explainer.fit(fitted_logistic_regression, *fake_dataset)
    explanation = explainer.explain_local(pd.DataFrame([[5, 0.1]]))
    assert len(explanation) == 1
    single_explanation = explanation[0]
    assert abs(single_explanation['real']) > abs(single_explanation['fake'])


def test_lime_explainer_multiple(fake_dataset, fitted_logistic_regression):
    explainer = LimeExplainer(class_names=['false', 'true'])
    explainer.fit(fitted_logistic_regression, *fake_dataset)
    _do_explainer_test(explainer)


def test_lime_explainer_array(fake_dataset, fitted_logistic_regression):
    explainer = LimeExplainer(class_names=['false', 'true'])
    explainer.fit(fitted_logistic_regression, fake_dataset[0].values, fake_dataset[1])
    _do_explainer_test(explainer, np.array([[5, 0.1], [95, -0.5]]), col_real='feature_0', col_fake='feature_1')


def test_lime_explainer_series(fake_dataset, fitted_logistic_regression):
    explainer = LimeExplainer(class_names=['false', 'true'])
    explainer.fit(fitted_logistic_regression, *fake_dataset)
    _do_explainer_test(explainer, pd.Series([5, 0.1], index=['real', 'fake']))


def test_lime_xgb(fake_dataset):
    model = XGBClassifier()
    x, y = fake_dataset
    model.fit(x.values, y)

    explainer = LimeExplainer()
    explainer.fit(model, *fake_dataset)
    _do_explainer_test(explainer)


def test_lime_nn(fake_dataset, fitted_neural_network):

    explainer = LimeExplainer(class_names=['false', 'true'])
    explainer.fit(fitted_neural_network, *fake_dataset)
    explanation = explainer.explain_local(pd.DataFrame([[5, 0.1], [95, -0.5]]))
    assert len(explanation) == 2



import operator
from typing import List, Optional, Dict, Callable, Tuple, Union

import pandas as pd
import numpy as np
import sklearn

from trelawney.base_explainer import BaseExplainer
import trelawney.tree_explainer
# import trelawney.logreg_explainer


class SurrogateExplainer(BaseExplainer):
    """
    A surrogate model is a substitution model used to explain the initial model. Therefore, substitution models are
    generally simpler than the initial ones. Here, we use single trees and logistic regressions as surrogates.
    """

    def __init__(self, surrogate_model: sklearn.base.BaseEstimator, class_names: Optional[List[str]] = None):
        if type(surrogate_model) not in [sklearn.tree.tree.DecisionTreeClassifier,
                                         sklearn.linear_model.base.LinearRegression]:
            raise NotImplementedError('SurrogateExplainer is only available for single trees (single_tree) and logistic'
                                      'regression (logistic_regression) at the time being.')
        self._surrogate = surrogate_model
        self._explainer = None
        self._x_train = None
        self._model_to_explain = None
        self._adequation_metric = None
        self._class_names = class_names

    def fit(self, model: sklearn.base.BaseEstimator, x_train: Union[pd.Series, pd.DataFrame, np.ndarray],
            y_train: pd.DataFrame):
        x_train = self._get_dataframe_from_mixed_input(x_train)
        self._model_to_explain = model
        self._x_train = x_train.values
        if type(self._surrogate) == sklearn.tree.tree.DecisionTreeClassifier:
            self._surrogate.fit(self._x_train, self._model_to_explain.predict(self._x_train))
            self._adequation_metric = sklearn.metrics.accuracy_score
            self._explainer = trelawney.tree_explainer.TreeExplainer(self._class_names).fit(self._surrogate, x_train, y_train)
        else:
            raise NotImplementedError
            # self._surrogate.fit(x_train, self._model_to_explain.predict_probas(x_train))
            # self._adequation_metric = sklearn.metrics.mean_squared_error
            # self._explainer = trelawney.logreg_explainer.LogRegExplainer().fit(self._surrogate)
        return self

    def adequation_score(self, metric: Union[Callable[[np.ndarray, np.ndarray], float], str] = 'auto', ):
        """
        returns an adequation score between the output of the surrogate and the output of the initial model based on
        the x_train set given.
        """
        if metric != 'auto':
            self._adequation_metric = metric
        return self._adequation_metric(self._model_to_explain.predict(self._x_train),
                                       self._surrogate.predict(self._x_train))

    def feature_importance(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                           n_cols: Optional[int] = None) -> Dict[str, float]:
        """
        returns a relative importance of each feature globally as a dict.
        :param x_explain: the dataset to explain on
        :param n_cols: the maximum number of features to return
        """
        x_explain = self._get_dataframe_from_mixed_input(x_explain)
        return self._explainer.feature_importance(x_explain, n_cols)

    def explain_local(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray], n_cols: Optional[int] = None, ) -> List[Dict[str, float]]:
        """
        returns local relative importance of features for a specific observation.
        :param x_explain: the dataset to explain on
        :param n_cols: the maximum number of features to return
        """
        x_explain = self._get_dataframe_from_mixed_input(x_explain)
        return self._explainer.explain_local(x_explain, n_cols)

    def plot_tree(self, out_path: str = './tree_viz'):
        """returns the colored plot of the decision tree and saves an Image in the wd."""
        if type(self._surrogate) != sklearn.tree.tree.DecisionTreeClassifier:
            raise TypeError('plot_tree is only available for single tree surrogate')
        return self._explainer.plot_tree(out_path=out_path)




"""
module that provides the base explainer class from which all future explainers will inherit
"""
import abc
import operator
from typing import List, Optional, Dict, Union

import sklearn
import numpy as np
import pandas as pd
import plotly.graph_objs as go
from trelawney.colors import BLUE, GREY


class BaseExplainer(abc.ABC):
    """
    the base explainer class. this is an abstract class so you will need to define some behaviors when implementing your
    new explainer. In order to do so, override:

    - the `fit` method that defines how (if needed) the explainer should be fited
    - the `feature_importance` method that extracts the relative importance of each feature on a dataset globally
    - the `explain_local` method that extracts the relative impact of each feature on the final decisionfor every sample
      in a dataset
    """

    def __init__(self):
        self._model_to_explain = None

    def fit(self, model: sklearn.base.BaseEstimator, x_train: Union[pd.Series, pd.DataFrame, np.ndarray],
            y_train: pd.DataFrame):
        """
        prepares the explainer by saving all the information it needs and fitting necessary models

        :param model: the TRAINED model the explainer will need to shed light on
        :param x_train: the dataset the model was trained on originally
        :param y_train: the target the model was trained on originally
        """
        self._model_to_explain = model

    @abc.abstractmethod
    def feature_importance(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                           n_cols: Optional[int] = None) -> Dict[str, float]:
        """
        returns a relative importance of each feature on the predictions of the model (the explainer was fitted on) for
        `x_explain` globally. The output will be a dict with the importance for each column/feature in `x_explain`
        (limited to n_cols)

        if some importance are negative, this means they are negatively correlated with the output and absolute value
        represents the relative importance

        :param x_explain: the dataset to explain on
        :param n_cols: the maximum number of features to return (ordered by importance)
        """
        pass

    @staticmethod
    def _filter_and_limit_dict(col_importance_dic: Dict[str, float], cols: List[str], n_cols: int):
        og_mvmt = sum(col_importance_dic.values())
        sorted_and_filtered = dict(sorted(
            filter(
                lambda col_and_importance: col_and_importance[0] in cols,
                col_importance_dic.items()
            ),
            key=lambda x: abs(x[1]),
            reverse=True
        )[:n_cols])
        sorted_and_filtered['rest'] = sorted_and_filtered.get('rest', 0.) + (og_mvmt - sum(sorted_and_filtered.values()))
        return sorted_and_filtered

    def filtered_feature_importance(self, x_explain: pd.DataFrame, cols: Optional[List[str]],
                                    n_cols: Optional[int] = None) -> Dict[str, float]:
        """same as `feature_importance` but applying a filter first (on the name of the column)"""
        return self._filter_and_limit_dict(self.feature_importance(x_explain), cols, n_cols)

    def graph_feature_importance(self, x_explain: pd.DataFrame, cols: Optional[List[str]] = None, n_cols: Optional[int] = None,
                                 irrelevant_cols: Optional[List[str]] = None):
        cols = cols or x_explain.columns.to_list()
        irrelevant_cols = irrelevant_cols or []
        feature_importance = self.filtered_feature_importance(x_explain, cols, n_cols)
        rest = feature_importance.pop('rest')
        sorted_feature_importance = sorted(feature_importance.items(), key=operator.itemgetter(1), reverse=True)
        colors = [BLUE if col not in irrelevant_cols else GREY
                  for col in map(operator.itemgetter(0), sorted_feature_importance)]
        colors.append(GREY)
        plot = go.Bar(x=list(map(operator.itemgetter(0), sorted_feature_importance)) + ['rest'],
                      y=list(map(operator.itemgetter(1), sorted_feature_importance)) + [rest],
                      marker_color=colors)
        return go.Figure(plot)

    @staticmethod
    def _get_dataframe_from_mixed_input(undecided_input):
        """transforms several accepted formats (for now DataFrames, Series and arrays) to data-frames"""
        if isinstance(undecided_input, pd.DataFrame):
            return undecided_input
        if isinstance(undecided_input, pd.Series):
            return pd.DataFrame(undecided_input).T
        if isinstance(undecided_input, np.ndarray) and len(undecided_input.shape) == 1:
            return pd.DataFrame(undecided_input.reshape(1, -1), columns=['feature_{}'.format(i)
                                                                         for i in range(undecided_input.shape[0])])
        if isinstance(undecided_input, np.ndarray) and len(undecided_input.shape) == 2:
            return pd.DataFrame(undecided_input, columns=['feature_{}'.format(i)
                                                          for i in range(undecided_input.shape[1])])
        if isinstance(undecided_input, np.ndarray):
            raise ValueError('arrays with more than two dimensions not supported')
        raise TypeError('unsupported type: {}'.format(type(undecided_input)))

    @abc.abstractmethod
    def explain_local(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                      n_cols: Optional[int] = None) -> List[Dict[str, float]]:
        """
        explains each individual predictions made on x_explain. BEWARE this is usually quite slow on large datasets

        :param x_explain: the samples to explain
        :param n_cols: the number of columns to limit the explanation to
        """
        if not isinstance(x_explain, (pd.DataFrame, pd.Series, np.ndarray)):
            raise TypeError('{} is not supported, please use dataframes, Series or arrays'.format(type(x_explain)))

    def explain_filtered_local(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray], cols: List[str],
                               n_cols: Optional[int] = None) -> List[Dict[str, float]]:
        """same as `explain_local` but applying a filter on each explanation on the features"""

        return [
            self._filter_and_limit_dict(sample_importance_dict, cols, n_cols)
            for sample_importance_dict in self.explain_local(x_explain)
        ]

    def graph_local_explanation(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                                cols: Optional[List[str]] = None, n_cols: Optional[int] = None,
                                info_values: Optional[Union[pd.DataFrame, pd.Series]] = None) -> go.Figure:
        """
        creates a waterfall plotly figure to represent the influance of each feature on the final decision for a single
        prediction of the model.

        You can filter the columns you want to see in your graph and limit the final number of columns you want to see.
        If you choose to do so the filter will be applied first and of those filtered columns at most `n_cols` will be
        kept

        :param x_explain: the example of the model this must be a dataframe with a single ow
        :param cols: the columns to keep if you want to filter (if None - default) all the columns will be kept
        :param n_cols: the number of columns to limit the graph to. (if None - default) all the columns will be kept

        :raises ValueError: if x_explain doesn't have the right shape
        """
        if x_explain.shape[0] != 1:
            raise ValueError('can only explain single observations, if you only have one sample, use reshape(1, -1)')

        info_values = x_explain if info_values is None else info_values

        info_values = self._get_dataframe_from_mixed_input(info_values)
        x_explain = self._get_dataframe_from_mixed_input(x_explain)
        # transforming the info values to a Series
        info_values = info_values.iloc[0, :]
        print(info_values)
        print(x_explain)

        # checking info columns and x_explain match
        if any(info_values.index != x_explain.columns):
            raise ValueError(
                'columns differ from x_explain ({}) and info_values({})'.format(x_explain.columns, info_values.index)
            )

        cols = cols or x_explain.columns.to_list()
        importance_dict = self.explain_filtered_local(x_explain, cols=cols, n_cols=n_cols)[0]

        output_value = self._model_to_explain.predict_proba(x_explain.values)[0, 1]
        start_value = output_value - sum(importance_dict.values())
        rest = importance_dict.pop('rest')

        sorted_importances = sorted(importance_dict.items(), key=lambda importance: abs(importance[1]), reverse=True)
        hovertext = ['start value',
                     *['{} = {}'.format(col_name, info_values[col_name]) for col_name, col_value in sorted_importances],
                     'rest', 'output_value = {}'.format(output_value)]
        fig = go.Figure(go.Waterfall(
            orientation="v",
            measure=['absolute', *['relative' for _ in importance_dict], 'relative', 'absolute'],
            y=[start_value, *map(operator.itemgetter(1), sorted_importances), rest, output_value],
            textposition="outside",
            #     text = ["+60", "+80", "", "-40", "-20", "Total"],
            x=['start_value', *map(operator.itemgetter(0), sorted_importances), 'rest', 'output_value'],
            connector={"line": {"color": GREY}},
            decreasing={"marker": {"color": '#DB643D'}},
            increasing={"marker": {"color": '#3DDC97'}},
            totals={"marker": {"color": BLUE}},
            hovertext=hovertext,
            hoverinfo='text+delta',
        ))
        fig.update_layout(
            title="explanation",
            showlegend=False
        )
        return fig



"""
Module that provides the LogRegExplainer class base on the BaseExplainer class
"""
import operator
from typing import Optional, List, Dict, Union

import pandas as pd
import numpy as np
import sklearn
import plotly.graph_objs as go
from trelawney.base_explainer import BaseExplainer
from trelawney.colors import BLUE, GREY


class LogRegExplainer(BaseExplainer):
    """
    The LogRegExplainer class is composed of 3 methods:
    - fit: get the right model
    - feature_importance (global interpretation)
    - graph_odds_ratio (visualisation of the ranking of the features, based on their odds ratio)
    """

    def __init__(self, class_names: Optional[List[str]] = None, categorical_features: Optional[List[str]] = None, ):
        """
        initialize class_names, categorical_features and model_to_explain
        """
        self.class_names = class_names
        self._model_to_explain = None
        self._feature_names = None

    def fit(self, model: sklearn.base.BaseEstimator, x_train: Union[pd.Series, pd.DataFrame, np.ndarray],
            y_train: pd.DataFrame):
        x_train = self._get_dataframe_from_mixed_input(x_train)
        self._model_to_explain = model
        self._feature_names = x_train.columns

    def feature_importance(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                           n_cols: Optional[int] = None) -> Dict[str, float]:
        """
        returns the absolute value (i.e. magnitude) of the coefficient of each feature as a dict.
        :param x_explain: the dataset to explain on
        :param n_cols: the maximum number of features to return
        """
        x_explain = self._get_dataframe_from_mixed_input(x_explain)
        res = dict(zip(x_explain.columns, np.abs(self._model_to_explain.coef_[0])))
        return res

    def _compute_odds_ratio(self):
        res = pd.DataFrame({'variables': self._feature_names,
                            'odds_ratio': np.exp(self._model_to_explain.coef_[0])}).set_index('variables')
        return res

    def graph_odds_ratio(self, n_cols: Optional[int] = 10, ascending: bool = False,
                         irrelevant_cols: Optional[List[str]] = None) -> pd.DataFrame:
        """
        returns a plot of the top k features, based on the magnitude of their odds ratio.
        :n_cols: number of features to plot
        :ascending: order of the ranking of the magnitude of the coefficients
        """
        irrelevant_cols = irrelevant_cols or []
        all_odds = self._compute_odds_ratio()
        odds_ratio_dict = dict(zip(all_odds.index, all_odds.odds_ratio))
        dataset_to_plot = sorted(odds_ratio_dict.items(), key=lambda x: abs(x[1]), reverse=not ascending)[:n_cols]

        colors = [BLUE if col not in irrelevant_cols else GREY
                  for col in map(operator.itemgetter(0), dataset_to_plot)]
        plot = go.Bar(x=list(map(operator.itemgetter(0), dataset_to_plot)),
                      y=list(map(operator.itemgetter(1), dataset_to_plot)),
                      marker_color=colors)
        fig = go.Figure(plot)
        fig.update_layout(
            title='Top {} variables, based on their odds ratio'.format(n_cols),
            showlegend=True
        )
        return fig

    def explain_local(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                      n_cols: Optional[int] = None) -> List[Dict[str, float]]:
        """
        returns local relative importance of features for a specific observation.
        :param x_explain: the dataset to explain on
        :param n_cols: the maximum number of features to return
        """
        raise NotImplementedError('no consensus on which values can explain the path followed by an observation')



import operator
from typing import List, Optional, Dict, Union

import pandas as pd
import numpy as np
import sklearn
from lime import lime_tabular
from tqdm import tqdm

from trelawney.base_explainer import BaseExplainer


class LimeExplainer(BaseExplainer):
    """
    Lime stands for local interpretable model-agnostic explanations and is a package based on
    `this article <https://www.arxiv.org/abs/1602.04938>`_. Lime will explain a single prediction of you model
    by crechariotsating a local approximation of your model around said prediction.'sphinx.ext.autodoc', 'sphinx.ext.viewcode']

    .. testsetup::

        >>> import pandas as pd
        >>> import numpy as np
        >>> from trelawney.lime_explainer import LimeExplainer
        >>> from sklearn.linear_model import LogisticRegression

    .. doctest::

        >>> X = pd.DataFrame([np.array(range(100)), np.random.normal(size=100).tolist()], index=['real', 'fake']).T
        >>> y = np.array(range(100)) > 50
        >>> # training the base model
        >>> model = LogisticRegression().fit(X, y)
        >>> # creating and fiting the explainer
        >>> explainer = LimeExplainer()
        >>> explainer.fit(model, X, y)
        <trelawney.lime_explainer.LimeExplainer object at ...>
        >>> # explaining observation
        >>> explanation =  explainer.explain_local(pd.DataFrame([[5, 0.1]]))[0]
        >>> abs(explanation['real']) > abs(explanation['fake'])
        True


    """

    def __init__(self, class_names: Optional[List[str]] = None, categorical_features: Optional[List[str]] = None, ):
        super().__init__()
        self._explainer = None
        if class_names is not None and len(class_names) != 2:
            raise NotImplementedError('Trelawney only handles binary classification case for now. PR welcome ;)')
        self.class_names = class_names
        self._output_len = None
        self.categorical_features = categorical_features

    def fit(self, model: sklearn.base.BaseEstimator, x_train: Union[pd.Series, pd.DataFrame, np.ndarray],
            y_train: pd.DataFrame, ):
        x_train = self._get_dataframe_from_mixed_input(x_train)
        super().fit(model, x_train, y_train)
        self._explainer = lime_tabular.LimeTabularExplainer(x_train.values, feature_names=x_train.columns,
                                                            class_names=self.class_names,
                                                            categorical_features=self.categorical_features,
                                                            discretize_continuous=True)
        return self

    def feature_importance(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                           n_cols: Optional[int] = None) -> Dict[str, float]:
        raise NotImplementedError('we are not sure global explaination is mathematically sound for LIME, it is still'
                                  ' debated, refer tp https://github.com/skanderkam/trelawney/issues/10')

    @staticmethod
    def _extract_col_from_explanation(col_explanation):
        is_left_term = len([x for x in col_explanation if x in ['<', '>']]) < 2
        if is_left_term:
            return col_explanation.split()[0]
        return col_explanation.split()[2]

    def explain_local(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                      n_cols: Optional[int] = None) -> List[Dict[str, float]]:
        x_explain = self._get_dataframe_from_mixed_input(x_explain)
        super().explain_local(x_explain)
        x_explain = self._get_dataframe_from_mixed_input(x_explain)
        n_cols = n_cols or len(x_explain.columns)
        res = []
        for individual_sample in tqdm(x_explain.iterrows()):
            individual_explanation = self._explainer.explain_instance(individual_sample[1],
                                                                      self._model_to_explain.predict_proba,
                                                                      num_features=x_explain.shape[1],
                                                                      top_labels=2)
            individual_explanation = sorted(individual_explanation.as_list(), key=operator.itemgetter(1), reverse=True)
            skewed_individual_explanation = {self._extract_col_from_explanation(col_name): col_importance
                                             for col_name, col_importance in individual_explanation[:n_cols]}
            rest = sum(map(operator.itemgetter(1), individual_explanation)) - sum(skewed_individual_explanation.values())
            skewed_individual_explanation['rest'] = rest
            res.append(skewed_individual_explanation)
        return res



from typing import List, Optional, Dict, Union

import pandas as pd
import shap as shap
import sklearn
import numpy as np
import operator
import keras.models
from keras.wrappers.scikit_learn import KerasClassifier
from sklearn.ensemble.forest import ForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.tree.tree import BaseDecisionTree
from tqdm import tqdm
from xgboost import XGBClassifier

from trelawney.base_explainer import BaseExplainer


class ShapExplainer(BaseExplainer):
    """
    SHAP is a `package <https://github.com/slundberg/shap>` to estimate Shapley values on an individual prediction.

    Shapley come from sport analitics by trying to devide the outcome of a match betweeen the players of team by
    comparing the estimated outcome of the match with all the combinaton of players. For model interpretation, we
    consider that all the features are the players and the probability of the classifier the outcome of the game
    """

    def __init__(self):
        super().__init__()
        self._explainer = None

    def _find_right_explainer(self, x_train):
        if isinstance(self._model_to_explain, LogisticRegression):
            return shap.LinearExplainer(self._model_to_explain, data=x_train.values)
        if isinstance(self._model_to_explain, (BaseDecisionTree, ForestClassifier, XGBClassifier)):
            return shap.TreeExplainer(self._model_to_explain)
        if isinstance(self._model_to_explain, KerasClassifier):
            return shap.DeepExplainer(self._model_to_explain.model, data=x_train.values)
        return shap.KernelExplainer(self._model_to_explain, data=x_train.values)

    def fit(self, model: sklearn.base.BaseEstimator, x_train: Union[pd.Series, pd.DataFrame, np.ndarray],
            y_train: pd.DataFrame):
        x_train = self._get_dataframe_from_mixed_input(x_train)
        super().fit(model, x_train, y_train)
        self._explainer = self._find_right_explainer(x_train)
        return self

    def _get_shap_values(self, x_explain):
        shap_values = self._explainer.shap_values(x_explain.values)
        if isinstance(shap_values, list):
            # if we have a list this means there is one output per class, we take the positive class
            # TODO class should be a parameter
            return shap_values[1]
        return shap_values

    def explain_local(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                      n_cols: Optional[int] = None) -> List[Dict[str, float]]:
        super().explain_local(x_explain)
        x_explain = self._get_dataframe_from_mixed_input(x_explain)
        shap_values = self._get_shap_values(x_explain)
        n_cols = n_cols or len(x_explain.columns)
        res = []
        for individual_sample in tqdm(range(len(x_explain))):

            # creating the contributions of all the features to a prediction
            individual_explanation = list(zip(x_explain.columns, shap_values[individual_sample].T))

            # limiting to n_cols
            total_mvmt = sum(map(operator.itemgetter(1), individual_explanation))
            skewed_individual_explanation = sorted(individual_explanation, key=operator.itemgetter(1),
                                                   reverse=True)[:n_cols]
            skewed_individual_explanation.append(
                ('rest', total_mvmt - sum(map(operator.itemgetter(1), skewed_individual_explanation)))
            )
            res.append(dict(skewed_individual_explanation))
        return res

    def feature_importance(self, x_explain: Union[pd.Series, pd.DataFrame, np.ndarray],
                           n_cols: Optional[int] = None) -> Dict[str, float]:
        x_explain = self._get_dataframe_from_mixed_input(x_explain)
        shap_values = self._get_shap_values(x_explain)
        shap_dict = dict(zip(x_explain.columns.to_list(), np.mean(np.abs(shap_values), axis=0).reshape(-1).tolist()))
        kept_shap_bar_cols = dict(sorted(shap_dict.items(), key=lambda x: abs(x[1]), reverse=True,)[:n_cols])
        return kept_shap_bar_cols



import pandas as pd
from sklearn.tree import DecisionTreeClassifier

from trelawney.surrogate_explainer import SurrogateExplainer


def test_surogate_explainer_single(fake_dataset, fitted_decision_tree):
    explainer = SurrogateExplainer(DecisionTreeClassifier(max_depth=3))
    explainer.fit(fitted_decision_tree, *fake_dataset)
    explanation = explainer.feature_importance(pd.DataFrame([[30, 0.1]], columns=['real', 'fake']))
    assert abs(explanation['real']) > abs(explanation['fake'])


def test_surogate_explainer_multiple(fake_dataset, fitted_decision_tree):
    explainer = SurrogateExplainer(DecisionTreeClassifier(max_depth=3))
    explainer.fit(fitted_decision_tree, *fake_dataset)
    explanation = explainer.feature_importance(pd.DataFrame([[5, 0.1], [95, -0.5]], columns=['real', 'fake']))
    assert abs(explanation['real']) > abs(explanation['fake'])
